import logging
from typing import List, Optional, Union

import pandas as pd
from anndata import AnnData

log = logging.getLogger(__name__)


def _get_counts(adata: AnnData, groupby, key, normalize) -> pd.DataFrame:
    if groupby is None:
        return adata.obs[[key]].apply(lambda s: s.value_counts(normalize)).T

    grouped = adata.obs.groupby(groupby)
    return grouped[key].apply(lambda s: s.value_counts(normalize)).unstack(level=-1)


def cell_type_ratios(
    adata: AnnData,
    groupby: Union[str, List[str], None] = None,
    normalize: bool = True,
    key: str = "scyan_pop",
    among: str = None,
) -> pd.DataFrame:
    """Count for each patient (or group) the number of cells for each population.

    Args:
        adata: An `AnnData` object.
        groupby: Key(s) of `adata.obs` used to create groups (e.g. the patient ID).
        normalize: If `False`, returns counts instead of percentages.
        key: Key of `adata.obs` containing the population names (or the values to count).
        among: Key of `adata.obs` containing the parent population name. For example, if 'T CD4 RM' is found in `adata.obs[key]`, then we may find something like 'T cell' in `adata.obs[among]`. Typically, if using hierarchical populations, you can provide `'scyan_pop_level'` with your level name.

    Returns:
        A DataFrame of counts (one row per group, one column per population).
    """
    normalize = among is not None or normalize
    column_suffix = "percentage" if normalize else "count"

    counts = _get_counts(adata, groupby, key, normalize)

    if among is None:
        counts.columns = [f"{name} {column_suffix}" for name in counts.columns]
        return counts

    parents_count = _get_counts(adata, groupby, among, normalize)

    df_parent = adata.obs.groupby(among)[key].apply(lambda s: s.value_counts()).unstack()
    assert (
        (df_parent > 0).sum(0) == 1
    ).all(), f"Each population from adata.obs['{key}'] should have one and only one parent population in adata.obs['{among}']"
    to_parent_dict = dict(df_parent.idxmax())

    counts /= parents_count[[to_parent_dict[pop] for pop in counts.columns]].values
    counts.columns = [
        f"{pop} {column_suffix} among {to_parent_dict[pop]}" for pop in counts.columns
    ]
    return counts


def mean_intensities(
    adata: AnnData,
    groupby: Union[str, List[str], None] = None,
    layer: Optional[str] = None,
    key: str = "scyan_pop",
    unstack_join: Optional[str] = " mean intensity on ",
    obsm: Optional[str] = None,
    obsm_names: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Compute the Mean Metal Intensity (MMI) or Mean Fluorescence Intensity (MFI) per population. If needed, mean intensities can be computed per group (e.g., per patient) by providing the `groupby` argument.

    Args:
        adata: An `AnnData` object.
        groupby: Key(s) of `adata.obs` used to create groups. For instance, `"id"` computes MMI per population for each ID. You can also provide something like `["group", "id"]` to get MMI per group, and per patient inside each group.
        layer: In which `adata.layers` we get expression intensities. By default, it uses `adata.X`.
        key: Key of `adata.obs` containing the population names.
        unstack_join: If `None`, keep the information grouped. Else, flattens the biomarkers into one series (or one row per group if `groupby` is a list) and uses `unstack_join` to join the names of the multi-level columns. For instance, `' expression on '` can be a good choice.
        obsm: In which `adata.obsm` we get expression intensities. By default, it uses `adata.X`. If not `None` then `obsm_names` is required too.
        obsm_names: Ordered list of names in `adata.obsm[obsm]` if `obsm` was provided.

    Returns:
        A DataFrame of MFI. If `groupby` was a list, it is a multi-index dataframe.
    """
    if groupby is None:
        groupby = [key]
    elif isinstance(groupby, str):
        groupby = [groupby, key]
    else:
        groupby = list(groupby) + [key]

    if obsm is not None:
        assert (
            layer is None
        ), "You must choose between 'obsm' and 'layer', do not use both."

        df = pd.DataFrame(data=adata.obsm[obsm], columns=obsm_names)
    else:
        df = adata.to_df(layer)

    for group in groupby:
        df[group] = adata.obs[group].values

    res = df.groupby(groupby).mean().dropna(how="all")

    if res.values.min() < 0:
        log.warn(
            "The minimum expression value is negative. Are you sure you are using unscaled values? If not, you can use 'scyan.preprocess.unscale' and save the unscaled result in a 'adata.layers' of your choice (then use this layer argument in the current function). If you know what you are doing, or if you use flow cytometry data, you can ignore this warning."
        )

    if unstack_join is None:
        return res

    res = res.unstack(level=-1)
    if isinstance(res, pd.Series):
        res.index = [unstack_join.join(row).strip() for row in res.index.values]
    else:
        res.columns = [unstack_join.join(col).strip() for col in res.columns.values]
    return res



from typing import Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch
from scipy.stats import norm

from .. import Scyan
from ..utils import _get_subset_indices
from .utils import check_population, plot_decorator


@torch.no_grad()
@plot_decorator()
def pops_expressions(
    model: Scyan,
    latent: bool = True,
    key: str = "scyan_pop",
    n_cells: Optional[int] = 200_000,
    vmax: float = 1.2,
    vmin: float = -1.2,
    cmap: Optional[str] = None,
    figsize: Tuple[float] = (10, 6),
    show: bool = True,
):
    """Heatmap that shows (latent or standardized) cell expressions for all populations.

    !!! note
        If using the latent space, it will only show the marker you provided to Scyan. Else, it shows every marker of the panel.

    Args:
        model: Scyan model.
        latent: If `True`, displays Scyan's latent expressions, else just the standardized expressions.
        key: Key to look for populations in `adata.obs`. By default, uses the model predictions.
        n_cells: Number of cells to be considered for the heatmap (to accelerate it when $N$ is very high). If `None`, consider all cells.
        vmax: Maximum value on the heatmap.
        vmax: Minimum value on the heatmap.
        cmap: Colormap name. By default, uses `"coolwarm"` if `latent`, else `"viridis"`.
        figsize: Pair `(width, height)` indicating the size of the figure.
        show: Whether or not to display the figure.
    """
    not_na = ~model.adata.obs[key].isna()
    indices = _get_subset_indices(not_na.sum(), n_cells)
    indices = np.where(not_na)[0][indices]

    x = model(indices).numpy(force=True) if latent else model.adata[indices].X
    columns = model.var_names if latent else model.adata.var_names

    df = pd.DataFrame(x, columns=columns, index=model.adata.obs.index[indices])
    df["Population"] = model.adata[indices].obs[key]

    if cmap is None:
        cmap = "coolwarm" if latent else "viridis"

    plt.figure(figsize=figsize)
    sns.heatmap(df.groupby("Population").mean(), vmax=vmax, vmin=vmin, cmap=cmap)
    plt.title(f"{'Latent' if latent else 'Standardized'} expressions grouped by {key}")


@torch.no_grad()
@plot_decorator()
def boxplot_expressions(
    model: Scyan,
    marker: str,
    latent: bool = False,
    key: str = "scyan_pop",
    ascending: bool = False,
    figsize: tuple = (10, 5),
    fliersize: float = 0.5,
    show: bool = True,
):
    """Boxplot of expressions for all population (for one marker).

    Args:
        model: Scyan model.
        marker: Name of the marker whose expressions are to be displayed.
        latent: If `True`, displays Scyan's latent expressions, else just the standardized expressions.
        key: Key to look for populations in `adata.obs`. By default, uses the model predictions.
        ascending: Whether to sort populations by ascending mean expressions.
        figsize: Figure size. Defaults to (10, 5).
        fliersize: Outlier dot size (see seaborn boxplot).
        show: Whether or not to display the figure.
    """
    if latent:
        assert (
            marker in model.var_names
        ), f"Marker {marker} was not used during the model training, impossible to use 'latent=True'."

        u = model()
        y = u[:, model.var_names.get_loc(marker)].numpy(force=True)
    else:
        y = model.adata.obs_vector(marker)

    df = pd.DataFrame({key: model.adata.obs[key], "y": y})
    order = df.groupby(key).mean().sort_values("y", ascending=ascending).index

    plt.figure(figsize=figsize)
    sns.boxplot(data=df, x=key, y="y", fliersize=fliersize, order=order)
    plt.xticks(rotation=90)
    plt.ylabel(f"{marker} {'latent' if latent else 'standardized'} expression")


@torch.no_grad()
@plot_decorator()
@check_population(one=True)
def pop_expressions(
    model: Scyan,
    population: str,
    key: str = "scyan_pop",
    max_value: float = 1.5,
    num_pieces: int = 100,
    radius: float = 0.05,
    figsize: Tuple[float] = (2, 6),
    show: bool = True,
):
    """Plot latent cell expressions for one population. Contrary to `scyan.plot.pops_expressions`, in displays expressions on a vertical bar, from `Neg` to `Pos`.

    Args:
        model: Scyan model.
        population: Name of one population to interpret. To be valid, the population name has to be in `adata.obs[key]`.
        key: Key to look for populations in `adata.obs`. By default, uses the model predictions.
        max_value: Maximum absolute latent value.
        num_pieces: Number of pieces to display the colorbar.
        radius: Radius used to chunk the colorbar. Increase this value if multiple names overlap.
        figsize: Pair `(width, height)` indicating the size of the figure.
        show: Whether or not to display the figure.
    """
    condition = model.adata.obs[key] == population
    u_mean = model(condition).mean(dim=0)
    values = u_mean.numpy(force=True).clip(-max_value, max_value)

    y = np.linspace(-max_value, max_value, num_pieces + 1)
    cmap = plt.get_cmap("RdBu")
    y_cmap = norm.pdf(np.abs(y) - 1, scale=model.hparams.prior_std)
    y_cmap = y_cmap - y_cmap.min()
    y_cmap = 0.5 - np.sign(y) * (y_cmap / y_cmap.max() / 2)
    colors = cmap(y_cmap).clip(0, 0.8)

    plt.figure(figsize=figsize, dpi=100)
    plt.vlines(np.zeros(num_pieces), y[:-1], y[1:], colors=colors, linewidth=5)
    plt.annotate("Pos", (-0.7, 1), fontsize=15)
    plt.annotate("Neg", (-0.7, -1), fontsize=15)

    for v in np.arange(-max_value, max_value, 2 * radius):
        labels = [
            label
            for value, label in zip(values, model.var_names)
            if abs(v - value) < radius
        ]
        if labels:
            plt.plot([0, 0.1], [v, v], "k")
            plt.annotate(", ".join(labels), (0.2, v - 0.03))

    plt.xlim([-1, 1])
    plt.axis("off")



from typing import Tuple

import pytorch_lightning as pl
import torch
from torch import Tensor, nn

from . import CouplingLayer


class RealNVP(pl.LightningModule):
    """Normalizing flow module (more specifically the RealNVP transformation $f_{\phi}$).

    Attributes:
        module (nn.Sequential): Sequence of [coupling layers][scyan.module.CouplingLayer].
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_size: int,
        n_hidden_layers: int,
        n_layers: int,
    ):
        """
        Args:
            input_size: Input size, i.e. number of markers + covariates
            hidden_size: MLP (`s` and `t`) hidden size.
            output_size: Output size, i.e. number of markers.
            n_hidden_layers: Number of hidden layers for the MLP (`s` and `t`).
            n_layers: Number of coupling layers.
        """
        super().__init__()
        self.module_list = nn.ModuleList(
            [
                CouplingLayer(
                    input_size,
                    hidden_size,
                    output_size,
                    n_hidden_layers,
                    self._mask(output_size, i),
                )
                for i in range(n_layers)
            ]
        )
        self.module = nn.Sequential(*self.module_list)

    def _mask(self, output_size: int, shift: int):
        return (((torch.arange(output_size) - shift) % 3) > 0).to(int)

    def forward(self, x: Tensor, covariates: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """Forward implementation, i.e. $f_{\phi}$.

        Args:
            x: Inputs of size $(B, M)$.
            covariates: Covariates of size $(B, M_c)$

        Returns:
            Tuple of (outputs, covariates, lod_det_jacobian sum)
        """
        return self.module((x, covariates, None))

    def inverse(self, u: Tensor, covariates: Tensor) -> Tensor:
        """Go through the RealNVP in reverse direction, i.e. $f_{\phi}^{-1}$.

        Args:
            u: Latent expressions of size $(B, M)$.
            covariates: Covariates of size $(B, M_c)$

        Returns:
            Outputs of size $(B, M)$.
        """
        for module in reversed(self.module_list):
            u = module.inverse(u, covariates)
        return u



import pytorch_lightning as pl
import torch
from torch import Tensor, distributions


class PriorDistribution(pl.LightningModule):
    """Prior distribution $U$"""

    def __init__(self, rho: Tensor, rho_mask: Tensor, prior_std: float, n_markers: int):
        """
        Args:
            rho: Tensor $\rho$ representing the knowledge table
            rho_mask: Mask telling where $\rho$ is NA
            prior_std: Standard deviation $\sigma$ for $H$.
            n_markers: Number of markers in the table.
        """
        super().__init__()
        self.prior_std = prior_std
        self.n_markers = n_markers

        self.register_buffer("rho", rho)
        self.register_buffer("rho_mask", rho_mask)
        self.register_buffer("loc", torch.zeros((n_markers)))
        self.register_buffer("cov", torch.eye((n_markers)) * self.prior_std**2)

        self.uniform = distributions.Uniform(-1, 1)
        self.normal = distributions.Normal(0, self.prior_std)

        self.compute_constant_terms()

    @property
    def prior_h(self) -> distributions.Distribution:
        """The distribution of $H$"""
        return distributions.MultivariateNormal(self.loc, self.cov)

    def compute_constant_terms(self) -> None:
        self.uniform_law_radius = 1 - self.prior_std

        _gamma = (
            self.uniform_law_radius
            / self.prior_std
            * torch.sqrt(2 / torch.tensor(torch.pi))
        )
        self.gamma = 1 / (1 + _gamma)

        na_constant_term = self.rho_mask.sum(dim=1) * torch.log(self.gamma)
        self.register_buffer("na_constant_term", na_constant_term)

    def difference_to_modes(self, u: Tensor) -> Tensor:
        """Difference between the latent variable $U$ and all the modes (one mode per population).

        Args:
            u: Latent variables tensor of size $(B, M)$.

        Returns:
            Tensor of size $(B, P, M)$ representing differences to all modes.
        """
        diff = u[:, None, :] - self.rho[None, ...]

        diff[:, self.rho_mask] = torch.clamp(
            diff[:, self.rho_mask].abs() - self.uniform_law_radius, min=0
        )  # Handling NA values

        return diff

    def log_prob_per_marker(self, u: Tensor) -> Tensor:
        """Log probability per marker and per population.

        Args:
            u: Latent variables tensor of size $(B, M)$.

        Returns:
            Log probabilities tensor of size $(B, P, M)$.
        """
        diff = self.difference_to_modes(u)  # size N x P x M

        return self.normal.log_prob(diff) + self.rho_mask * torch.log(self.gamma)

    def log_prob(self, u: Tensor) -> Tensor:
        """Log probability per population.

        Args:
            u: Latent variables tensor of size $(B, M)$.

        Returns:
            Log probabilities tensor of size $(B, P)$.
        """
        diff = self.difference_to_modes(u)  # size N x P x M

        return self.prior_h.log_prob(diff) + self.na_constant_term

    def sample(self, z: Tensor) -> Tensor:
        """Sampling latent cell-marker expressions.

        Args:
            z: Tensor of population indices.

        Returns:
            Latent expressions, i.e. a tensor of size $(len(Z), M)$.
        """
        (n_samples,) = z.shape

        e = self.rho[z] + self.rho_mask[z] * self.uniform.sample(
            (n_samples, self.n_markers)
        )
        h = self.prior_h.sample((n_samples,))

        return e + h



from collections import defaultdict

import matplotlib.pyplot as plt
import pandas as pd

from .. import Scyan
from .utils import plot_decorator


@plot_decorator()
def pops_hierarchy(model: Scyan, figsize: tuple = (18, 5), show: bool = True) -> None:
    """Plot populations as a tree, where each level corresponds to more detailed populations. To run this function, your knowledge table need to contain at least one population 'level' (see [this tutorial](../../tutorials/usage/#working-with-hierarchical-populations)), and you need to install `graphviz`.

    Args:
        model: Scyan model.
        figsize: Matplotlib figure size.
        show: Whether or not to display the figure.
    """
    import networkx as nx
    from networkx.drawing.nx_agraph import graphviz_layout

    table = model.table

    assert isinstance(
        table.index, pd.MultiIndex
    ), "To plot population hierarchy, you need a MultiIndex DataFrame. See the documentation for more details."

    root = "All populations"

    G = nx.DiGraph()
    G.add_node(root)

    def add_nodes(table, indices, level, parent=root):
        if level == -1:
            return

        index = table.index.get_level_values(level)
        dict_indices = defaultdict(list)
        for i in indices:
            dict_indices[index[i]].append(i)

        for name, indices in dict_indices.items():
            if not name == parent:
                G.add_node(name)
                G.add_edge(parent, name)
            add_nodes(table, indices, level - 1, name)

    add_nodes(
        table,
        range(len(table)),
        table.index.nlevels - 1,
    )

    plt.figure(figsize=figsize)
    pos = graphviz_layout(G, prog="dot")
    nx.draw_networkx(G, pos, with_labels=False, arrows=False, node_size=0)
    for node, (x, y) in pos.items():
        plt.text(
            x,
            y,
            node,
            ha="center",
            va="center",
            rotation=90,
            bbox=dict(facecolor="wheat", edgecolor="black", boxstyle="round,pad=0.5"),
        )

    plt.grid(False)
    plt.box(False)



from typing import List, Optional, Union

import numpy as np
import pandas as pd
import seaborn as sns
from anndata import AnnData

from .. import Scyan
from ..utils import _get_subset_indices, _has_umap, _subset
from ._scanpy_plot import scanpy_pl_umap
from .utils import check_population, get_palette_others, plot_decorator, select_markers


@plot_decorator(adata=True)
@check_population(return_list=True)
def scatter(
    adata: AnnData,
    population: Union[str, List[str], None],
    markers: Optional[List[str]] = None,
    n_markers: Optional[int] = 3,
    key: str = "scyan_pop",
    max_obs: int = 2000,
    s: float = 1.0,
    show: bool = True,
) -> None:
    """Display marker expressions on 2D scatter plots with colors per population.
    One scatter plot is displayed for each pair of markers.

    Args:
        adata: An `anndata` object.
        population: One population, or a list of population to be colored, or `None`. If not `None`, the population name(s) has to be in `adata.obs[key]`.
        markers: List of markers to plot. If `None`, the list is chosen automatically.
        n_markers: Number of markers to choose automatically if `markers is None`.
        key: Key to look for populations in `adata.obs`. By default, uses the model predictions.
        max_obs: Maximum number of cells per population to be displayed. If population is None, then this number is multiplied by 10.
        s: Dot marker size.
        show: Whether or not to display the figure.
    """
    markers = select_markers(adata, markers, n_markers, key, population)

    if population is None:
        indices = _get_subset_indices(adata.n_obs, max_obs * 10)
        data = adata[indices, markers].to_df()
        g = sns.PairGrid(data, corner=True)
        g.map_offdiag(sns.scatterplot, s=s)
        g.map_diag(sns.histplot)
        g.add_legend()
        return

    data = adata[:, markers].to_df()
    keys = adata.obs[key].astype(str)
    data["Population"] = np.where(~np.isin(keys, population), "Others", keys)

    pops = list(population) + ["Others"]
    if max_obs is not None:
        groups = data.groupby("Population").groups
        data = data.loc[[i for pop in pops[::-1] for i in _subset(groups[pop], max_obs)]]

    palette = get_palette_others(data, "Population")

    g = sns.PairGrid(data, hue="Population", corner=True, palette=palette, hue_order=pops)
    g.map_offdiag(sns.scatterplot, s=s)
    g.map_diag(sns.histplot)
    g.add_legend()


@plot_decorator(adata=True)
def umap(
    adata: AnnData,
    color: Union[str, List[str]] = None,
    vmax: Union[str, float] = "p95",
    vmin: Union[str, float] = "p05",
    show: bool = True,
    **scanpy_kwargs: int,
):
    """Plot a UMAP using scanpy.

    !!! note
        If you trained your UMAP with [scyan.tools.umap][] on a subset of cells, it will only display the desired subset of cells.

    Args:
        adata: An `anndata` object.
        color: Marker(s) or `obs` name(s) to color. It can be either just one string, or a list (it will plot one UMAP per element in the list).
        vmax: `scanpy.pl.umap` vmax argument.
        vmin: `scanpy.pl.umap` vmin argument.
        show: Whether or not to display the figure.
        **scanpy_kwargs: Optional kwargs provided to `scanpy.pl.umap`.
    """
    assert isinstance(
        adata, AnnData
    ), f"umap first argument has to be an AnnData object. Received type {type(adata)}."

    has_umap = _has_umap(adata)
    if not has_umap.all():
        adata = adata[has_umap]

    if color is None:
        return scanpy_pl_umap(adata, **scanpy_kwargs)

    return scanpy_pl_umap(adata, color=color, vmax=vmax, vmin=vmin, **scanpy_kwargs)


def pop_level(
    model: Scyan,
    group_name: str,
    level_name: str = "level",
    key: str = "scyan_pop",
    **scanpy_kwargs: int,
) -> None:
    """Plot all subpopulations of a group at a certain level on a UMAP (according to the populations levels provided in the knowledge table).

    Args:
        model: Scyan model.
        group_name: The group to look at among the populations of the selected level.
        level_name: Name of the column of the knowledge table containing the names of the grouped populations.
        key: Key of `adata.obs` to access the model predictions.
    """
    adata = model.adata
    table = model.table

    assert isinstance(
        table.index, pd.MultiIndex
    ), "To use this function, you need a MultiIndex DataFrame, see: https://mics-lab.github.io/scyan/tutorials/usage/#working-with-hierarchical-populations"

    level_names = table.index.names[1:]
    assert (
        level_name in level_names
    ), f"Level '{level_name}' unknown. Choose one of: {level_names}"

    base_pops = table.index.get_level_values(0)
    group_pops = table.index.get_level_values(level_name)
    assert (
        group_name in group_pops
    ), f"Invalid group name '{group_name}'. It has to be one of: {', '.join(group_pops)}."

    valid_populations = [
        pop for pop, group in zip(base_pops, group_pops) if group == group_name
    ]
    key_name = f"{key}_one_level"
    adata.obs[key_name] = pd.Categorical(
        [pop if pop in valid_populations else np.nan for pop in adata.obs[key]]
    )
    umap(
        adata,
        color=key_name,
        title=f"Among {group_name}",
        na_in_legend=False,
        **scanpy_kwargs,
    )



import logging
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from anndata import AnnData
from umap import UMAP

from ..utils import _check_is_processed, _get_subset_indices, _has_umap

log = logging.getLogger(__name__)


def _leiden(
    adata: AnnData, resolution: float, key_added: str, n_neighbors: int = 15
) -> None:
    try:
        import leidenalg
    except:
        raise ImportError(
            """To run leiden, you need to have 'leidenalg' installed. You can install the population discovery extra with "pip install 'scyan[discovery]'", or directly install leidenalg with "conda install -c conda-forge leidenalg"."""
        )

    import igraph as ig
    from sklearn.neighbors import kneighbors_graph

    if not "knn_graph" in adata.obsp:
        adata.obsp["knn_graph"] = kneighbors_graph(
            adata.X, n_neighbors=n_neighbors, metric="euclidean", include_self=False
        )

    # TODO (improvement): add weights according to euclidean distance
    graph = ig.Graph.Weighted_Adjacency(adata.obsp["knn_graph"], mode="DIRECTED")

    partition = leidenalg.find_partition(
        graph,
        leidenalg.RBConfigurationVertexPartition,
        resolution_parameter=resolution,
    )
    adata.obs[key_added] = pd.Categorical([str(x) for x in partition.membership])


def subcluster(
    adata: AnnData,
    population: str,
    markers: Optional[List[str]] = None,
    key: str = "scyan_pop",
    resolution: float = 0.2,
    size_ratio_th: float = 0.02,
    min_cells_th: int = 200,
    n_cells: int = 100_000,
) -> None:
    """Create sub-clusters among a given populations, and filters small clusters according to (i) a minimum number of cells and (ii) a minimum ratio of cells.
    !!! info
        After having run this method, you can analyze the results with [scyan.plot.umap][] and [scyan.plot.pops_expressions][].

    Args:
        adata: An `anndata` object.
        population: Name of the population to target (one of `adata.obs[key]`).
        markers: Optional list of markers used to create subclusters. By default, uses the complete panel.
        key: Key to look for population in `adata.obs`. By default, uses the model predictions, but you can also choose a population level (if any), or other observations.
        resolution: Resolution used for leiden clustering. Higher resolution leads to more clusters.
        size_ratio_th: (Only used if `population` is `None`): Minimum ratio of cells to be considered as a significant cluster (compared to the parent cluster).
        min_cells_th: (Only used if `population` is `None`): Minimum number of cells to be considered as a significant cluster.
        n_cells: Number of cells to be considered for the subclustering (to accelerate it when $N$ is very high). If `None`, consider all cells.
    """
    leiden_key = f"leiden_{resolution}_{population}"
    subcluster_key = f"scyan_subcluster_{population}"
    condition = adata.obs[key] == population
    markers = list(adata.var_names if markers is None else markers)

    if leiden_key in adata.obs and adata.uns.get(leiden_key, []) == markers:
        log.info(
            "Found leiden labels with the same resolution. Skipping leiden clustering."
        )
        indices = np.where(~adata.obs[leiden_key].isna())[0]
        adata_sub = adata[indices, markers].copy()
    else:
        has_umap = _has_umap(adata)
        if has_umap.all() or condition.sum() <= n_cells:
            indices = _get_subset_indices(condition.sum(), n_cells)
            indices = np.where(condition)[0][indices]
        else:
            indices = _get_subset_indices((condition & has_umap).sum(), n_cells)
            indices = np.where(condition & has_umap)[0][indices]

            k = len(indices)
            if k < n_cells:
                indices2 = _get_subset_indices((condition & ~has_umap).sum(), n_cells - k)
                indices2 = np.where(condition & ~has_umap)[0][indices2]
                indices = np.concatenate([indices, indices2])

        adata_sub = adata[indices, markers].copy()

        _leiden(adata_sub, resolution, leiden_key)

    series = pd.Series(index=np.arange(adata.n_obs), dtype=str)
    series[indices] = adata_sub.obs[leiden_key].values
    adata.obs[leiden_key] = series.values
    adata.obs[leiden_key] = adata.obs[leiden_key].astype("category")

    counts = adata_sub.obs[leiden_key].value_counts()
    remove = counts < max(counts.sum() * size_ratio_th, min_cells_th)

    assert (
        not remove.all()
    ), "All subclusters where filtered. Consider updating size_ratio_th and/or min_cells_th."

    adata_sub.obs.loc[
        np.isin(adata_sub.obs[leiden_key], remove[remove].index), leiden_key
    ] = np.nan

    series = pd.Series(index=np.arange(adata.n_obs), dtype=str)
    series[indices] = adata_sub.obs[leiden_key].values
    adata.obs[subcluster_key] = series.values
    adata.obs[subcluster_key] = adata.obs[subcluster_key].astype("category")

    adata.uns[leiden_key] = markers
    log.info(
        f"Subclusters created, you can now use:\n   - scyan.plot.umap(adata, color='{subcluster_key}') to show the clusters\n   - scyan.plot.pops_expressions(model, key='{subcluster_key}') to plot their expressions"
    )


def umap(
    adata: AnnData,
    markers: Optional[List[str]] = None,
    obsm: Optional[str] = None,
    n_cells: Optional[int] = 200_000,
    min_dist: float = 0.5,
    obsm_key: str = "X_umap",
    filter: Optional[Tuple] = None,
    **umap_kwargs: int,
) -> UMAP:
    """Run a [UMAP](https://umap-learn.readthedocs.io/en/latest/) on a specific set of markers (or all markers by default). It can be useful to show differences that are due to some markers of interest, instead of using the whole panel.

    !!! info

        This function returns a UMAP reducer. You can reuse it with `reducer.transform(...)` or save it with [scyan.data.add][].

    !!! note

        To actually plot the UMAP, use [scyan.plot.umap][].

    Args:
        adata: An `anndata` object.
        markers: List marker names. By default, use all the panel markers, i.e., `adata.var_names`.
        obsm: Name of the obsm to consider to train the UMAP. By default, uses `adata.X`.
        n_cells: Number of cells to be considered for the UMAP (to accelerate it when $N$ is very high). If `None`, consider all cells.
        min_dist: Min dist UMAP parameter.
        obsm_key: Key for `adata.obsm` to add the embedding.
        filter: Optional tuple `(key, value)` used to train the UMAP on a set of cells that satisfies a constraint. `key` is the key of `adata.obs` to consider, and `value` the value the cells need to have.
        **umap_kwargs: Optional kwargs to provide to the `UMAP` initialization.

    Returns:
        UMAP reducer.
    """
    reducer = UMAP(min_dist=min_dist, **umap_kwargs)

    if markers is None:
        markers = adata.var_names

    adata.obsm[obsm_key] = np.zeros((adata.n_obs, 2))
    indices = _get_subset_indices(adata.n_obs, n_cells)
    adata_view = adata[indices, markers]
    X = adata_view.X if obsm is None else adata_view.obsm[obsm]

    _check_is_processed(X)

    log.info("Fitting UMAP...")
    if filter is None:
        embedding = reducer.fit_transform(X)
    else:
        key, value = filter
        reducer.fit(X[adata[indices].obs[key] == value])
        log.info("Transforming...")
        embedding = reducer.transform(X)

    adata.obsm[obsm_key][indices] = embedding

    return reducer



import importlib.metadata
import logging

import pytorch_lightning as pl

from .model import Scyan
from ._io import read_fcs, read_csv, write_fcs, write_csv
from . import data, plot, tools, utils
from . import preprocess

__version__ = importlib.metadata.version("scyan")

logging.getLogger().setLevel(logging.INFO)
plot.reset_plot_settings()
pl.seed_everything(0)



import colorsys
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd


class GroupPalette:
    def __init__(self, alpha_l: float, step_l: float, alpha_s: float, step_s: float):
        self.alpha_l = alpha_l
        self.alpha_s = alpha_s
        self.step_l = step_l
        self.step_s = step_s

        self.k_l = 1 + int((1 - 2 * alpha_l) // step_l)
        self.k_s = 1 + int((1 - 2 * alpha_s) // step_s)
        self.max_size = self.k_l * self.k_s

    def get_ls(self, i: int, j: int):
        l = 1 - self.alpha_l - i * self.step_l
        s = 1 - self.alpha_s - j * self.step_s
        return l, s

    def get_block(self, h: float, size: int) -> List[Tuple[float]]:
        assert (
            size <= self.max_size
        ), f"Too many categories ({size}): can define at most {self.max_size} colors per group. Consider lowering these arguments: step_l, step_s, alpha_l, alpha_s."

        indices = [(i % self.k_l, i // self.k_l) for i in range(size)]
        indices = sorted(indices)

        return [colorsys.hls_to_rgb(h, *self.get_ls(i, j)) for i, j in indices]

    def __call__(self, sizes: List[int], hue_shift: float) -> List:
        hues = sorted(range(len(sizes)), key=lambda x: [x % 3, x])
        hues = np.array(hues) / len(sizes) + hue_shift

        return [self.get_block(h, size) for h, size in zip(hues, sizes)]


def palette_level(
    table: pd.DataFrame,
    population_index: Union[int, str] = 0,
    level_index: Union[int, str] = 1,
    hue_shift: float = 0.4,
    alpha_l: float = 0.25,
    step_l: float = 0.15,
    alpha_s: float = 0.3,
    step_s: float = 0.4,
) -> Dict[str, Tuple[float]]:
    """Computes a color palette that in grouped by the hierarchical main populations. It improves the UMAP readability when many populations are defined.

    !!! info
        Once such a color palette is defined, you can use it for plotting. For instance, try `scyan.plot.umap(adata, color="scyan_pop", palette=palette)`, where `palette` is the one you created with this function.

    Args:
        table: Knowledge table provided to Scyan. It must be a multi-index DataFrame.
        population_index: Index or name of the level in `table.index` storing the low-level/children population names.
        level_index: Index or name of the level in `table.index` storing the main population names.
        hue_shift: Shift the hue values. The value must be a float in `[0, 1]`.
        alpha_l: Lower it to have a larger lightness range of colors.
        step_l: Increase it to have more distinct colors (in term of lightness).
        alpha_s: Lower it to have a larger saturation range of colors.
        step_s: Increase it to have more distinct colors (in term of saturation).

    Returns:
        A dictionnary whose keys are population names and values are RGB colors.
    """
    assert isinstance(
        table.index, pd.MultiIndex
    ), f"The provided table has no multi-index. To work with hierarchical populations, consider reading https://mics-lab.github.io/scyan/tutorials/usage/#working-with-hierarchical-populations"

    pops = table.index.get_level_values(population_index).values
    level = table.index.get_level_values(level_index)
    level_counts = level.value_counts()

    group_palette = GroupPalette(alpha_l, step_l, alpha_s, step_s)
    color_groups = group_palette(level_counts.values, hue_shift)

    block_indices = [level_counts.index.get_loc(pop) for pop in level]
    s = pd.Series(level)
    inner_block_indices = s.groupby(s).cumcount().values

    return {
        pop: list(color_groups[block_index][inner_index])
        for pop, block_index, inner_index in zip(pops, block_indices, inner_block_indices)
    }



import logging
from typing import List, Optional

import flowutils
import numpy as np
import scipy
from anndata import AnnData

from .utils import _subset

log = logging.getLogger(__name__)


def auto_logicle_transform(
    adata: AnnData, q: float = 0.05, m: float = 4.5, quantile_clip: Optional[float] = 1e-5
) -> None:
    """[Auto-logicle transformation](https://pubmed.ncbi.nlm.nih.gov/16604519/) implementation.
    We recommend it for flow cytometry or spectral flow cytometry data.

    Args:
        adata: An `anndata` object.
        q: See logicle article. Defaults to 0.05.
        m: See logicle article. Defaults to 4.5.
    """
    adata.uns["scyan_logicle"] = {}
    markers_failed = []

    for marker in adata.var_names:
        column = adata[:, marker].X.toarray().flatten()

        w = 0
        t = column.max()
        negative_values = column[column < 0]

        if negative_values.size:
            threshold = np.quantile(negative_values, 0.25) - 1.5 * scipy.stats.iqr(
                negative_values
            )
            negative_values = negative_values[negative_values >= threshold]

            if negative_values.size:
                r = 1e-8 + np.quantile(negative_values, q)
                if 10**m * abs(r) > t:
                    w = (m - np.log10(t / abs(r))) / 2

        if not w or w > 2:
            markers_failed.append(marker)
            w, t = 1, 5e5

        column = flowutils.transforms.logicle(column, None, t=t, m=m, w=w)
        adata.uns["scyan_logicle"][marker] = [t, m, w]

        if quantile_clip is None:
            adata[:, marker] = column
        else:
            adata[:, marker] = column.clip(np.quantile(column, quantile_clip))

    if markers_failed:
        log.warn(
            f"Auto logicle transformation failed for the following markers (logicle was used instead): {', '.join(markers_failed)}.\nIt can happen when expressions are all positive or all negative."
        )


def _logicle_inverse_one(adata: AnnData, obsm: Optional[str], marker: str) -> np.ndarray:
    column = adata[:, marker].X if obsm is None else adata[:, marker].obsm[obsm]
    column = column.flatten()
    return flowutils.transforms.logicle_inverse(
        column, None, *adata.uns["scyan_logicle"][marker]
    )


def asinh_transform(adata: AnnData, translation: float = 0, cofactor: float = 5) -> None:
    """Asinh transformation for cell-expressions: $asinh((x - translation)/cofactor)$.

    Args:
        adata: An `anndata` object.
        translation: Constant substracted to the marker expression before division by the cofactor.
        cofactor: Scaling factor before computing the asinh.
    """
    adata.uns["scyan_asinh"] = [translation, cofactor]
    adata.X = np.arcsinh((adata.X - translation) / cofactor)


def inverse_transform(
    adata: AnnData,
    obsm: Optional[str] = None,
    obsm_names: Optional[List[str]] = None,
    transformation: Optional[str] = None,
) -> np.ndarray:
    """Inverses the transformation function, i.e. either [scyan.preprocess.auto_logicle_transform][] or [scyan.preprocess.asinh_transform][]. It requires to have run have of these before.

    !!! note
        If you scaled your data, the complete inverse consists in running [scyan.preprocess.unscale][] first, and then this function.

    Args:
        adata: An `anndata` object.
        obsm: Name of the anndata obsm to consider. If `None`, use `adata.X`.
        obsm_names: Names of the ordered markers from obsm. It is required if obsm is not `None`, if there are less markers than in `adata.X`, and if the transformation to reverse is `logicle`. Usually, it corresponds to `model.var_names`.
        transformation: Name of the transformation to inverse: one of `['logicle', 'asinh', None]`. By default, it chooses automatically depending on which transformation was previously run.

    Returns:
        Inverse transformed expressions array of shape $(N, M)$.
    """
    if transformation is None:
        transformation = "asinh" if "scyan_asinh" in adata.uns else None
        transformation = "logicle" if "scyan_logicle" in adata.uns else transformation
        if transformation is None:
            raise ValueError(
                "No transformation to inverse: you need to run 'asinh_transform' or 'auto_logicle_transform' before to inverse it."
            )

    if transformation == "logicle":
        log.info("Performing inverse logicle transform")
        assert (
            "scyan_logicle" in adata.uns
        ), "You need to run 'auto_logicle_transform' before to inverse it."

        if obsm is None:
            obsm_names = adata.var_names
        elif obsm_names is None:
            assert (
                adata.obsm[obsm].shape[1] != adata.n_vars
            ), f"When the number of var in adata.obsm['{obsm}'] is not `adata.n_vars`, use `obs_names`"
            obsm_names = adata.var_names

        return np.stack(
            [_logicle_inverse_one(adata, obsm, marker) for marker in obsm_names],
            axis=1,
        )

    if transformation == "asinh":
        log.info("Performing inverse asinh transform")
        assert (
            "scyan_asinh" in adata.uns
        ), "You need to run 'asinh_transform' before to inverse it."

        X = adata.X if obsm is None else adata.obsm[obsm]
        translation, cofactor = adata.uns["scyan_asinh"]

        return np.sinh(X) * cofactor + translation

    raise NameError(
        f"Parameter 'transformation' has to be 'logicle' or 'asinh'. Found {transformation}."
    )


def scale(adata: AnnData, max_value: float = 10, center: Optional[bool] = None) -> None:
    """Tranforms the data such as (i) `std=1`, and (ii) either `0` is sent to `-1` (for CyTOF data) or `means=0` (for flow or spectral flow data); except if `center` is set (which overwrites the default behavior).

    Args:
        adata: An `anndata` object.
        max_value: Clip to this value after scaling.
        center: If `None`, data is only centered for spectral or flow cytometry data (recommended), else, it is centered or not according to the value given.
    """
    stds = adata.X.std(axis=0)
    adata.uns["scyan_scaling_stds"] = stds

    if not center or (center is None and "scyan_asinh" in adata.uns):
        adata.X = (adata.X / stds - 1).clip(-max_value, max_value)
    else:
        means = adata.X.mean(axis=0)
        adata.X = ((adata.X - means) / stds).clip(-max_value, max_value)
        adata.uns["scyan_scaling_means"] = means


def unscale(
    adata: AnnData, obsm: Optional[str] = None, obsm_names: Optional[List[str]] = None
) -> np.ndarray:
    """Reverse standardisation. It requires to have run [scyan.preprocess.scale][] before.

    Args:
        adata: An `anndata` object.
        obsm: Name of the adata obsm to consider. If `None`, use `adata.X`.
        obsm_names: Names of the ordered markers from obsm. It is required if obsm is not `None`, and if there are less markers than in `adata.X`. Usually, it corresponds to `model.var_names`.

    Returns:
        Unscaled numpy array of shape $(N, M)$.
    """
    assert (
        "scyan_scaling_stds" in adata.uns
    ), "It seems you haven't run 'scyan.preprocess.scale' before."

    X = adata.X if obsm is None else adata.obsm[obsm]
    stds = adata.uns["scyan_scaling_stds"]

    if obsm is not None and X.shape[1] != adata.n_vars:
        assert (
            obsm_names is not None
        ), f"Found {X.shape[1]} markers in adata.obsm['{obsm}'], but 'adata' has {adata.n_vars} vars. Please use the 'obsm_names' argument to provide the ordered names of the markers used in adata.obsm['{obsm}']."

        indices = [adata.var_names.get_loc(marker) for marker in obsm_names]
        stds = stds[indices]

    if "scyan_scaling_means" in adata.uns:
        return adata.uns["scyan_scaling_means"] + stds * X

    return (X + 1) * stds


def subsample(adata: AnnData, n_obs: int) -> AnnData:
    """Subsample an `AnnData` object (cells will be chosen randomly).

    Args:
        adata: `AnnData` object.
        n_obs: Number of cell to keep.

    Returns:
        A view `AnnData` object with `n_obs` cells.
    """
    indices = _subset(np.arange(adata.n_obs), n_obs)
    return adata[indices]



import matplotlib.pyplot as plt
import numpy as np
from anndata import AnnData

from ..utils import _has_umap


class _SelectFromCollection:
    """From https://matplotlib.org/stable/gallery/widgets/polygon_selector_demo.html"""

    def __init__(self, ax, collection, alpha_other=0.3):
        from matplotlib.widgets import PolygonSelector

        self.canvas = ax.figure.canvas
        self.collection = collection
        self.alpha_other = alpha_other

        self.xys = collection.get_offsets()
        self.n_obs = len(self.xys)

        self.fc = collection.get_facecolors()
        self.fc = np.tile(self.fc, (self.n_obs, 1))

        self.poly = PolygonSelector(ax, self.onselect, draw_bounding_box=True)
        self.ind = []

    def onselect(self, verts):
        from matplotlib.path import Path

        path = Path(verts)
        self.ind = np.nonzero(path.contains_points(self.xys))[0]
        self.fc[:, -1] = self.alpha_other
        self.fc[self.ind, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()

    def disconnect(self):
        self.poly.disconnect_events()
        # TODO: understand why it crashes
        # self.fc[:, -1] = 1
        # self.collection.set_facecolors(self.fc)
        # self.canvas.draw_idle()


class PolygonGatingUMAP:
    """Class used to select cells on a UMAP using polygons.

    !!! note

        We recommend using it on Jupyter Notebooks. To be able to select the cells, you should first run `%matplotlib tk` on a blank jupyter cell. After the selection, you can run `%matplotlib inline` to retrieve the default behavior.

    ```py
    # Usage example (to be run on a jupyter notebook) (`%matplotlib tk` is required for the cell selection)
    >>> %matplotlib tk
    >>> selector = scyan.tools.PolygonGatingUMAP(adata)
    >>> selector.select()         # select the cells
    >>> selector.save_selection() # save the selected cells in adata.obs
    >>> %matplotlib inline        # to retrieve the default behavior
    ```
    """

    def __init__(self, adata: AnnData) -> None:
        """
        Args:
            adata: An `anndata` object.
        """
        self.adata = adata
        self.has_umap = _has_umap(adata)
        self.x_umap = self.adata.obsm["X_umap"]

    def select(self, s: float = 0.05) -> None:
        """Open a UMAP plot on which you can draw a polygon to select cells.

        Args:
            s: Size of the cells on the plot.
        """
        _, ax = plt.subplots()

        pts = ax.scatter(
            self.x_umap[self.has_umap, 0],
            self.x_umap[self.has_umap, 1],
            marker=".",
            rasterized=True,
            s=s,
        )

        self.selector = _SelectFromCollection(ax, pts)

        print(
            f"Enclose cells within a polygon. Helper:\n    - Click on the plot to add a polygon vertex\n    - Press the 'esc' key to start a new polygon\n    - Try holding the 'ctrl' key to move a single vertex\n    - Once the polygon is finished and overlaid in red, you can close the window"
        )
        plt.show()

    def save_selection(self, key_added: str = "scyan_selected"):
        """Save the selected cells in `adata.obs[key_added]`.

        Args:
            key_added: Column name used to save the selected cells in `adata.obs`.
        """
        self.adata.obs[key_added] = "unselected"
        col_index = self.adata.obs.columns.get_loc(key_added)
        self.adata.obs.iloc[
            np.where(self.has_umap)[0][self.selector.ind], col_index
        ] = "selected"
        self.adata.obs[key_added] = self.adata.obs[key_added].astype("category")

        self.selector.disconnect()
        print(
            f"Selected {len(self.selector.ind)} cells and saved the selection in adata.obs['{key_added}']"
        )
